
/* Queue Manager:  */

#ifdef __USAGE
%C	[options] &

Where [options] can be:
          -g         Make queue a global network wide queue.
#endif

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <sys/psinfo.h>
#include <sys/timers.h>
#include <sys/kernel.h>
#include <sys/proxy.h>
#include <sys/name.h>
#include <sys/sys_msg.h>
#include <unistd.h>

/*
#define DEBUG
*/

/*
 * The user will interface to this mailbox administrator using a small
 * set of library routines of the form.
 *
 *		int queue_wait(id, wait)
 *		int id;
 *		int wait_type;
 *			{
 *			msg.msg_type = QUEUE_WAIT;
 *			msg.msg_queue_id = id;
 *			msg.msg_queue_wait = wait;
 *			send(queue_admin_tid, &msg, &msg, 4);
 *			return(msg.msg_type);
 *			}
 *
 * These routines are in the file queuelib.c
 */

/*
 * This is an example of the code required to implement a mailbox
 * system under QNX.
 * It should be very easy to customize this code to any application. 
 * It clearly illustrates the ease with which capabilities can be added
 * to QNX. Capabilities which require no change to the kernel yet add what
 * is perceived as kernel functions.
 *
 * Note that the QNX messaging performs the function of time ordered access
 * to this code WITHOUT the need for explicit semaphoring.  This mailbox
 * administrator process can be run on any node in the network and processes
 * on ANY machine may use it freely.  There is no need for special code to
 * implement networking.
 */

#define MAX_QUEUES		50    /* maximum number of queues this admin can have */
#define MAX_QUEUE_SIZE	2048  /* maximum size of any queue entry data */

#define QHDR_SIZE   	15    /* the size of the queuehdr structure in bytes */
#define QNAME_SIZE  	16    /* the maximum length of a queue name in bytes */

/*
 * Incoming message types
 */

#define QUEUE_OPEN			1
#define QUEUE_CLOSE			2
#define QUEUE_KILL			3
#define QUEUE_POST			4
#define QUEUE_READ			5
#define QUEUE_WAIT			6
#define QUEUE_INFO			7
#define QUEUE_EXIT			8
#define QUEUE_PURGE			9
#define QUEUE_WAIT_SIG		10
#define QUEUE_LAST_STATUS	11

/*
 * Return message codes
 */

#define OK				     0		
#define QSTAT_WAITM			-1
#define QSTAT_WAITW			-2
#define QSTAT_WAITE			-3
#define QSTAT_WAITR			-4
#define QSTAT_WAITS			-5
#define QSTAT_WAITT			-6

#define QSTAT_BADMSG		-9
#define QSTAT_OPEN_MODE		-10
#define QSTAT_OPEN_PERMS	-11
#define QSTAT_CREATE		-12
#define QSTAT_MEM			-13
#define QSTAT_EXIST			-14
#define QSTAT_WINDOW		-15
#define QSTAT_REPLY			-16
#define QSTAT_NAME			-17
#define QSTAT_FIND			-18
#define QSTAT_SLOT			-19
#define QSTAT_SEND			-20
#define QSTAT_MSGSIZE		-21
#define QSTAT_INTERNAL		-22

/*
 * Wait type flag definitions
 */

#define WAIT_MSG	0x01
#define WAIT_WRITER	0x02
#define WAIT_EMPTY	0x04
#define WAIT_READER	0x08
#define WAIT_WINDOW	0x10

#define WAIT_READ	0x40
#define WAIT_TIMER	0x80


/*
 * All messages received have this structure
 */

struct msg_entry {
	signed char msg_type;
	char msg_queue_id;
	int msg_queue_wait;
	int msg_queue_size;
	signed char msg_queue_priority;
	pid_t msg_queue_proxy;
	char msg_queue_data[MAX_QUEUE_SIZE];
	} ;

struct queue_info {			/* Overlayed on msg_queue_data[..] */
	pid_t queue_info_tids[20];
	unsigned queue_info_nmsgs;
	char queue_info_name[QNAME_SIZE+1];
	} ;

/*
 * A table of information indexed by a mailbox id which is a small INT
 * which is returned by the OPEN request.
 */

struct queue_table_entry {
	struct queue_entry *queue_list;	/* Waiting msgs */
	struct queue_entry *queue_last;	/* Last msgs in the list */
	struct wait_entry *wait_list;	/* Waiting for something */
	pid_t  queue_tid;				/* Process which created the mailbox */
	unsigned queue_nmsgs;			/* Number of waiting msgs */
	unsigned queue_window_size;		/* Window size */
	unsigned queue_flags;			/* Flags on the mailbox */
	signed char queue_open_count;	/* Number of opens */
	signed char queue_open_rcount;	/* Number of opens for r on this queue */
	signed char queue_open_wcount;	/* Number of opens for w on this queue */
	char queue_name[QNAME_SIZE+1];	/* Name used by QUEUE_OPEN */
	} ;

struct open_entry {
	struct open_entry *next_open;	/* next open by same task */
	int open_id;					/* queue_id of this open */
	char open_mode;					/* r/w flags */
	};

struct queue_entry {
	struct queue_entry *next_queue;	/* Next waiting msg */
	int queue_priority;				/* Priority of this message */
	unsigned queue_size;			/* Size in bytes of queue_data */
	char queue_data[1];				/* Variable length field */
	} ;

struct wait_entry {
	struct wait_entry *next_wait;	/* Next waiting task entry */
	struct wait_entry *next_twait;	/* Next waiting timer entry */
	struct queue_table_entry *wait_queue_ptr;
	pid_t  wait_pid;				/* Task id of waiting task */
	unsigned wait_size;				/* Max number of chars to read */
	char wait_type;					/* One flag bit for each class of wait */
	char wait_time;					/* # seconds to wait. */
	pid_t wait_proxy;				/* if not 0 signal this instead of reply */
	} ;

pid_t proxy;
int timerid;			    					/* Timer id for timed waits */
struct wait_entry *timer_queue = NULL;			/* Link list of timer waits */
struct open_entry *task_open_table[256];		/* Link list of opens */
struct open_entry *this_op;						/* matched open */

struct queue_table_entry queue_table[MAX_QUEUES];	/* Indexed by queue_id */
pid_t pid;	     								/* Process id of last QNX msg */
struct msg_entry msg;							/* The message buffer */
char status_save[256];							/* Cause of last signal */

int queue_open( void );
int queue_kill( void );
int queue_purge( void );
int queue_close( void );
int queue_post( void );
int queue_wait( int wait, pid_t wait_proxy );
int queue_last_status( void );
int queue_read( int wait );
int queue_info( void );
void check_waiting_tasks( struct queue_table_entry *tp, char mask );
struct queue_table_entry *ptr_to_queue( void );
int add_wait( struct queue_table_entry *tp, unsigned wait_type, pid_t wait_proxy);
void close_all( void );
void unlink_open( struct open_entry *entry, int id );
void unlink_wait( struct wait_entry **list, struct wait_entry *member );
void unlink_twait( struct wait_entry **list, struct wait_entry *member );
void check_timer( void );

#define min(a, b) (((a) < (b)) ? (a) : (b))


main(int argc, char **argv)
	{
	extern void queue_exit();
	int global, i;
	struct itimercb timercb;
    unsigned opt;
    char qname[25];
    struct {
        struct _sysmsg_hdr      hdr;
    } *deathmsg = &msg;          
    int pflags;

	global = 0;

	while((opt=getopt(argc,argv,"g"))!=-1) {
		switch(opt) {
 			case 'g': /* global flag */
                      global = 1;
                      break;
			default:  /* invalid option */
                      exit(EXIT_FAILURE);	
                      break;
		    }
	    }

	/*
	 * Attach to a proxy so that the timer can kick timed waits.
	 */

	if((proxy = qnx_proxy_attach(0, 0, 0, -1)) == -1) {
		fprintf(stderr, "Unable to attach a proxy: %s.\n",strerror(errno));
		exit(-1);
		}

	/*
	 * Register a name so processes can send us messages.
	 */
    if ( global ) 
       sprintf(qname,"/queue");
    else
       sprintf(qname,"queue");
	if((qnx_name_attach(0, qname)) == -1) {
		fprintf(stderr, "Unable to attach name %s: %s\n",qname,strerror(errno));
		exit(-1);
		}

	timercb.itcb_event.evt_value = proxy;
	if((timerid = mktimer(TIMEOFDAY, _TNOTIFY_PROXY, &timercb)) == -1) {
		fprintf(stderr, "Unable to attach a timer: %s.\n",strerror(errno));
		exit(-1);
		}

	signal(SIGINT, &queue_exit);

	for ( i = 0; i < 256; i++ )
		task_open_table[ i ] = NULL;
	memset( queue_table, 0, sizeof queue_table );
	memset( status_save, 0, sizeof status_save );

    pflags = _PPF_INFORM | _PPF_PRIORITY_REC;
	if(qnx_pflags(~0L, pflags, 0, 0)) {
		fprintf(stderr, "\nUnable to become informed admin: %s.\n", strerror(errno));
		exit(-1);
		}

	/*
	 * Now we block waiting for a message.
	 */

	for(;;) {
getmessage:
		pid = Receive(0, &msg, sizeof(msg));

#ifdef DEBUG
printf("Got message from %d. msg.msgtype=%d, msg.queueid=%d\n",
        pid, msg.msg_type, msg.msg_queue_id);
printf("deathmsg-hdr-msgtype=%d, deathmsg-hdr-subtype=%d\n",
        deathmsg->hdr.type, deathmsg->hdr.subtype);
#endif

		if(pid == proxy) {
			check_timer();
			goto getmessage;
			}

   /* look for death messages first */
   if ( deathmsg->hdr.type    == _SYSMSG
     && deathmsg->hdr.subtype == _SYSMSG_SUBTYPE_DEATH ) {
		Reply(pid, 0, 0);
	  	if ( task_open_table[pid & 0xff] )
           close_all();
        goto getmessage;
        }
		/*
		 * A return of OK indicates that there is no need to reply.
		 */
		switch ( msg.msg_type ) {
			case QUEUE_OPEN:
				msg.msg_type = queue_open();
				break;
	
			case QUEUE_POST:
				msg.msg_type = queue_post();
				break;
	
			case QUEUE_WAIT:
				msg.msg_type = queue_wait( msg.msg_queue_wait, 0 );
				break;
				
			case QUEUE_WAIT_SIG:
				msg.msg_type = queue_wait( msg.msg_queue_wait,
										   msg.msg_queue_proxy );
				break;
				
			case QUEUE_LAST_STATUS:
				msg.msg_type = queue_last_status();
				break;
				
			case QUEUE_READ:
				msg.msg_type = queue_read( msg.msg_queue_wait );
				break;
	
			case QUEUE_INFO:
				msg.msg_type = queue_info();
				break;
	
			case QUEUE_CLOSE:
				msg.msg_type = queue_close();
				break;
	
			case QUEUE_KILL:
				msg.msg_type = queue_kill();
				break;
	
			case QUEUE_PURGE:
				msg.msg_type = queue_purge();
				break;
	
			case QUEUE_EXIT:
				queue_exit();	/* Does not return */
				break;
	
		    default:
			    msg.msg_type = QSTAT_BADMSG;
			}
		    if(msg.msg_type != OK)
			    Reply(pid, &msg, 1);
     }
}


/*
 * Open a mailbox. If the mailbox
 * does not exist a new one is created if "c" option is provided.
 */

int queue_open( void )
{
	struct queue_table_entry *tp;
	char *p2, *p1;
	struct open_entry *op, *new_op;
	struct queue_table_entry *new_tp = NULL;
	unsigned mode = 0;
    #ifdef DEBUG
    int i;
    #endif
	static char letters[] = "rwcbhpt RW";	/* DO NOT change this order */

	/*
	 * Decode the open mode letters.
	 */

	for ( p1 = msg.msg_queue_data; *p1 ; ++p1 ) {
		for ( p2 = letters ; *p2 ; ++p2 )
			if ( *p1 == *p2 ) {
				mode |= 1 << (p2 - letters);
				break;
				}
		if ( *p2 == '\0' )
			return( QSTAT_OPEN_MODE );
		}
    p1++;   /* get past the null separator */
#ifdef DEBUG
for ( i=0; msg.msg_queue_data[i]; i++ )
   printf("qd[%d]=%c ",i,msg.msg_queue_data[i]);
printf("p1=%s\n",p1);
#endif

	/*
	 * Look for an existing mailbox first.
	 */

	for ( tp = &queue_table[0] ; tp < &queue_table[MAX_QUEUES] ; ++tp )
		if ( tp->queue_name[0] != '\0' ) {
			if ( strcmp( tp->queue_name, p1 ) == 0 ) {
				/* Check permission to open */
				if ( (mode & ~tp->queue_flags) & 0x03 )
					return ( QSTAT_OPEN_PERMS );
				goto found;
				}
			}
		else if ( new_tp == NULL )
			new_tp = tp;
				
	tp = new_tp;
	if ( (mode & 0x04) == 0 || tp == NULL )	/* Can't create mailbox */
		return ( QSTAT_CREATE );


	strcpy( tp->queue_name, p1 );		/* Create new mailbox */
#ifdef DEBUG
printf("opening new queue :%s\n",tp->queue_name);
#endif
	tp->queue_flags = (mode & ~0x0303) | ((mode >> 8) & 0x03);
	tp->queue_nmsgs = 0;
	tp->queue_open_count = tp->queue_open_rcount = tp->queue_open_wcount = 0;
	tp->queue_window_size = msg.msg_queue_size;
	tp->queue_list = NULL;
	tp->queue_last = NULL;	
	tp->wait_list  = NULL;

found:
#ifdef DEBUG
printf("opening existing queue %s\n",tp->queue_name);
#endif
	if ( (new_op = malloc( sizeof( *new_op ) )) == NULL ) {
		if ( tp == new_tp )
			tp->queue_name[0] = '\0';
		return ( QSTAT_MEM );
		}

	if ( (op = task_open_table[ pid & 0xff ]) == NULL )
		task_open_table[ pid & 0xff ] = new_op;
	else {
		while ( op->next_open != NULL )
			op = op->next_open;
		op->next_open = new_op;
		}

	new_op->open_id = tp - &queue_table[0];
	new_op->open_mode = mode;
	new_op->next_open = NULL;

	msg.msg_queue_id = tp - &queue_table[0];
	msg.msg_type = OK;
	Reply( pid, &msg, 4 );

	++tp->queue_open_count;
	if ( mode & 0x02 )
		++tp->queue_open_wcount;
	if ( mode & 0x01 )
		++tp->queue_open_rcount;
	return ( 0 );	
}


int queue_kill( void )
{
	struct open_entry *op1, *op2;
	struct queue_table_entry *tp;
	struct wait_entry *wp, *wp2;
	struct queue_entry *mp, *mp2;
	unsigned i;

	if ( (tp = ptr_to_queue()) == NULL )
		return ( QSTAT_EXIST );

	for ( i = 1 ; i < 256 ; ++i ) {
		op1 = task_open_table[ i ];
		while ( op1 != NULL ) {
			if ( msg.msg_queue_id == op1->open_id ) {
				op2 = op1->next_open;
				unlink_open( op1, i );
				free( op1 );
				op1 = op2;
				}
			else
				op1 = op1->next_open;
			}
		}

	wp = tp->wait_list;
	tp->wait_list = NULL;
	while ( wp != NULL ) {
		msg.msg_type = QSTAT_EXIST;
		if ( wp->wait_proxy )  {
			status_save[ wp->wait_pid & 0xff ] = msg.msg_type;
			Trigger( wp->wait_proxy );	/* Trigger rather than reply */
			}
		else
			Reply( wp->wait_pid, &msg, 1 );	/* Reply to all waiting processes */
		if ( wp->wait_type & WAIT_TIMER )
			unlink_twait( &timer_queue, wp );
		wp2 = wp->next_wait;
		free( wp );
		wp = wp2;
		}

	mp = tp->queue_list;
	while ( mp != NULL ) {
		mp2 = mp->next_queue;
		free( mp );
		mp = mp2;
		}

	tp->queue_name[0] = '\0';

	msg.msg_type = OK;
	Reply( pid, &msg, 1 );

	return ( OK );
}



int queue_purge( void )
{
	struct queue_table_entry *tp;
	struct queue_entry *mp, *mp1;

	if ( (tp = ptr_to_queue()) == NULL  ||
								(this_op->open_mode & 0x01) == 0 )
		return ( QSTAT_EXIST );

	for ( mp = tp->queue_list; mp != NULL; mp = mp1 ) {
		mp1 = mp->next_queue;
		free( mp );
		}
	tp->queue_list = NULL;
	tp->queue_last = NULL;		/* KWH */
	tp->queue_nmsgs = 0;

	check_waiting_tasks( tp, WAIT_EMPTY );

	msg.msg_type = OK;
	Reply( pid, &msg, 1 );

	return ( OK );
}



int queue_close( void )
{
	struct queue_table_entry *tp;
	struct wait_entry *wp1, *wp2;
	unsigned mode = 0;

	if ( (tp = ptr_to_queue()) == NULL )
		return ( QSTAT_EXIST );

	/* Remove all my wait entries */
	wp2 = NULL;
	wp1 = tp->wait_list;
	while ( NULL != wp1 ) {
		if ( wp1->wait_pid == pid ) {
			if ( wp1->wait_type & WAIT_TIMER )
				unlink_twait( &timer_queue, wp1 );
			if ( wp1 == tp->wait_list )
				tp->wait_list = wp1->next_wait;
			else
				wp2->next_wait = wp1->next_wait;
			free ( wp1 );
			}
		else
			wp2 = wp1;

		wp1 = (NULL == wp2) ? tp->wait_list : wp2->next_wait;
		}

	--tp->queue_open_count;
	if ( this_op->open_mode & 0x01 )
		if ( --tp->queue_open_rcount == 0 )
			mode |= WAIT_READER;

	if ( this_op->open_mode & 0x02 )
		if ( --tp->queue_open_wcount == 0 )
			mode |= WAIT_WRITER;

	if ( tp->queue_open_count < 0 || tp->queue_open_rcount < 0 ||
											tp->queue_open_wcount < 0 ) {
		fprintf( stderr, "QUEUE internal close error.\n" );
		queue_exit();
		}

	if ( tp->queue_flags & 0x0020 )
		/* "p" - permanent queue */
		goto close;

	if ((tp->queue_flags & 0x0010) &&
					(tp->queue_list != NULL || tp->queue_open_count != 0) )
		/* "h" - hold open as long as there are messages or opens */
		goto close;

	/* else the queue is temporary */
	if ( !(this_op->open_mode & 0x04) )
		/* not the creating task - don't remove the queue */
		goto close;

	queue_kill();

	tp->queue_list = NULL;	
	tp->queue_last = NULL;	
	tp->wait_list = NULL;	

	goto done;

close:
	if ( mode )
		check_waiting_tasks( tp, mode );
	unlink_open( this_op, pid & 0x00ff );
	free( this_op );

done:
	msg.msg_type = OK;
	Reply( pid, &msg, 1 );

	return ( 0 );
}


		
int queue_post( void )
{
	struct queue_table_entry *tp;
	struct queue_entry *mp1, *mp2, *new_mp;

	if ( (tp = ptr_to_queue()) == NULL  ||
						(this_op->open_mode & 0x02) == 0 )
		return ( QSTAT_EXIST );

	if ( tp->queue_window_size && tp->queue_nmsgs >= tp->queue_window_size )
		return ( QSTAT_WINDOW );

	/*
	 * Add to the linked list of mailbox entries.
	 * The message is inserted ahead of any entries of higher
	 * priority, but after messages of the same priority .
	 */

	if (msg.msg_queue_priority == -1)
	{
		mp2 = tp->queue_last;
		mp1 = NULL;
	}
	else
	{
		mp2 = NULL;	/* "trailing" pointer */
		for ( mp1 = tp->queue_list; mp1 != NULL; 
				mp2 = mp1, mp1 = mp1->next_queue )
		{
			if ( mp1->queue_priority > msg.msg_queue_priority )
				break;
		}
	}

	new_mp = malloc( sizeof( struct queue_entry ) + msg.msg_queue_size );
	if ( new_mp == NULL )
		return ( QSTAT_MEM );

	tp->queue_last = new_mp;	

	new_mp->next_queue = mp1;
	if ( mp2 == NULL )
		/* first message in queue */
		tp->queue_list = new_mp;
	else
		mp2->next_queue = new_mp;
	new_mp->queue_size = msg.msg_queue_size;
	new_mp->queue_priority = msg.msg_queue_priority ;
	memcpy( new_mp->queue_data, msg.msg_queue_data, msg.msg_queue_size );
	++tp->queue_nmsgs;

#ifdef DEBUG
printf("added data %s\n",msg.msg_queue_data);
#endif
	msg.msg_type = OK;
	Reply( pid, &msg, 1 );
	
	check_waiting_tasks( tp, WAIT_MSG | WAIT_READ );

	return ( OK );
}



int queue_wait( int wait, pid_t wait_proxy )
{
	struct queue_table_entry *tp;
	char bits = wait & 0xff;
	int i, rc, wait_bits = wait & 0xff00;

	if ( (tp = ptr_to_queue()) == NULL )
		return ( QSTAT_EXIST );

	if ( bits == 0 ) {
		msg.msg_type = (tp->queue_list == NULL) ? QSTAT_WAITE : QSTAT_WAITM;
		Reply( pid, &msg, 1 );
		}

	msg.msg_type = OK;
	for ( i = 0 ; bits ; ++i, bits >>= 1 ) {
		if ( (bits & 0x01) == 0 )
			continue;
		switch ( i ) {
			case 0:
				if ( tp->queue_list == NULL )
					wait_bits |= WAIT_MSG;
				break;
			case 1:
				if ( tp->queue_open_wcount )
					wait_bits |= WAIT_WRITER;
				break;
			case 2:
				if ( tp->queue_list != NULL )
					wait_bits |= WAIT_EMPTY;
				break;
			case 3:
				if ( tp->queue_open_rcount )
					wait_bits |= WAIT_READER;
				break;
			case 4:
				if ( tp->queue_window_size  &&
								tp->queue_nmsgs >= tp->queue_window_size )
					wait_bits |= WAIT_WINDOW;
				break;
			default:
				return ( QSTAT_INTERNAL );
			}
		}

	/*
	 * If the caller requested a signal rather than blocking, reply
	 * now.
	 */
	if ( wait_bits & 0x00ff )  {
		if ( ( (rc = add_wait( tp, wait_bits, wait_proxy )) == OK ) &&
														(wait_proxy != 0) ) {
			msg.msg_type = OK;
			Reply( pid, &msg, 1 );
			}
		return ( rc );
		}

	msg.msg_type = (tp->queue_list == NULL) ? QSTAT_WAITE : QSTAT_WAITM;
	Reply( pid, &msg, 1 );

	/*
	*	Signal here because the wait type was already satisfied and there
	*	was a proxy to signal
	*/
	if ( (wait_proxy != 0) && (wait & 0x00ff) )
		Trigger ( wait_proxy );

	return ( 0 );
}



int queue_last_status( void )
{
	if ( (msg.msg_type = status_save[pid & 0xff]) == OK )
		Reply( pid, &msg, 1 );

	return ( msg.msg_type );
}



int queue_read( int wait )
{
	struct queue_table_entry *tp;
	struct queue_entry *mp;
	int wait_bits, n;

	if ( (tp = ptr_to_queue()) == NULL  ||
							(this_op->open_mode & 0x01) == 0 )
		return ( QSTAT_EXIST );

#ifdef DEBUG
    printf("got read request. waittype=%x\n",wait);
#endif
 
	if ( (mp = tp->queue_list) == NULL ) {	/* No waiting msg */
		wait_bits = wait & 0xff00;
		if ( wait & WAIT_MSG )
			wait_bits |= WAIT_READ;
		if ( wait & WAIT_WRITER )
			if ( tp->queue_open_wcount != 0 )
				wait_bits |= WAIT_WRITER;
			else
				return ( QSTAT_WAITW );
		if ( wait_bits == 0 )
			return ( QSTAT_WAITE );
		return ( add_wait( tp, wait_bits, 0 ) );
		}

	msg.msg_queue_size = min( msg.msg_queue_size, mp->queue_size );
	n = mp->queue_size - msg.msg_queue_size;	/* Extra data */
	if ( (tp->queue_flags & 0x08) == 0  &&  n > 0 )
		return ( QSTAT_MSGSIZE );

	msg.msg_type = OK;
	memcpy( msg.msg_queue_data, mp->queue_data, msg.msg_queue_size );
#ifdef DEBUG
printf("read returns %s\n",msg.msg_queue_data);
#endif

	if ( Reply( pid, &msg, sizeof(msg) - (MAX_QUEUE_SIZE - mp->queue_size) )
																	== 0 ) {
		if ( (tp->queue_flags & 0x08)  &&  n > 0 ) {
			memcpy( mp->queue_data, mp->queue_data + msg.msg_queue_size, n );
			mp->queue_size = n;
			return ( OK );
			}
		tp->queue_list = mp->next_queue;		/* unlink the message */
		if ( --tp->queue_nmsgs <= tp->queue_window_size )
			check_waiting_tasks( tp, WAIT_WINDOW );
		if ( tp->queue_nmsgs == 0 )
			check_waiting_tasks( tp, WAIT_EMPTY );
		free( mp );

		if (tp->queue_list == NULL)
			tp->queue_last = NULL; 

		return ( OK );
		}

	return ( QSTAT_REPLY );
}



/*
 * This call can be used by tasks to return information
 * on the names of the mailboxes and the status of messages
 * and tasks waiting.  This is VERY useful when debugging
 * a real time system.  It requires that you write a
 * simple QNX utility to send the required INFO messages
 * and format and print the data returned.
 */

int queue_info( void )
{
	struct queue_table_entry *tp;
	struct wait_entry *wp;
	struct queue_info entry;
	int i;

	/*
	 * Start scan from passed mailbox id
	 */
	for ( ;; ++msg.msg_queue_id ) {
		if ( msg.msg_queue_id >= MAX_QUEUES )
			return ( QSTAT_EXIST );
		if ( (tp = &queue_table[msg.msg_queue_id])->queue_name[0] )
			break;
		}

	for ( i = 0, wp = tp->wait_list ; wp ; ++i, wp = wp->next_wait )
		entry.queue_info_tids[i] = wp->wait_pid;

	entry.queue_info_tids[i] = 0;
	entry.queue_info_nmsgs = tp->queue_nmsgs;
	strcpy( entry.queue_info_name, tp->queue_name );
	memcpy( msg.msg_queue_data, &entry, sizeof( struct queue_info ) );

	msg.msg_type = OK;
	Reply( pid, &msg, QHDR_SIZE + sizeof( struct queue_info ) );
	
	return ( OK );
}

/*
 * Check for tasks waiting for something.
 */

void check_waiting_tasks( struct queue_table_entry *tp, char mask )
{
	struct wait_entry *wp1, *wp2;
	pid_t tid_save = pid, stat;

	wp2 = NULL;			/* "trailing" pointer */
	wp1 = tp->wait_list;
	while (  wp1 != NULL ) {
		if ( wp1->wait_type & mask ) {
			if ( wp1->wait_type & WAIT_READ ) {	/* Handle as a special case */
				msg.msg_queue_size = wp1->wait_size;
				pid = wp1->wait_pid;
				if ( wp1->wait_type & WAIT_TIMER )
					unlink_twait( &timer_queue, wp1 );
				if( mask & WAIT_WRITER ) {
					msg.msg_type = QSTAT_WAITW;
					if ( wp1->wait_proxy )  {
						status_save[ wp1->wait_pid & 0xff ] = msg.msg_type;
						Trigger( wp1->wait_proxy );
						}
					else
						Reply( wp1->wait_pid, &msg, 1 );
					}
				else {
					if ( (stat = queue_read( 0 )) == OK ) {
						msg.msg_type = stat;
						if ( wp1->wait_proxy )  {
							status_save[ wp1->wait_pid & 0xff ] = msg.msg_type;
							Trigger( wp1->wait_proxy );
							}
						else
							Reply( wp1->wait_pid, &msg, 1 );
						}
					}
				}
			else {
				if ( mask & WAIT_MSG )
					msg.msg_type = QSTAT_WAITM;
				else if ( mask & WAIT_EMPTY )
					msg.msg_type = QSTAT_WAITE;
				else if ( mask & WAIT_WINDOW )
					msg.msg_type = QSTAT_WAITS;
				else if ( mask & WAIT_WRITER )
					msg.msg_type = QSTAT_WAITW;
				else if ( mask & WAIT_READER )
					msg.msg_type = QSTAT_WAITR;

				if ( wp1->wait_proxy )  {
					status_save[ wp1->wait_pid & 0xff ] = msg.msg_type;
					Trigger( wp1->wait_proxy );
					}
				else
					Reply( wp1->wait_pid, &msg, 1 );
				if ( wp1->wait_type & WAIT_TIMER )
					unlink_twait( &timer_queue, wp1 );
				}
			if ( wp1 == tp->wait_list )
				tp->wait_list = wp1->next_wait;
			else
				wp2->next_wait = wp1->next_wait;

			free( wp1 );
			}
		else
			wp2 = wp1;

		wp1 = (NULL == wp2) ? tp->wait_list : wp2->next_wait;
		}

	pid = tid_save;
}



/*
 * Return a pointer to a mailbox table entry. The mailbox id is checked
 * and we make sure the mailbox is open (ie: queue_name non-null).
 */

struct queue_table_entry *ptr_to_queue( void )
{
	struct open_entry *op1;
	struct queue_table_entry *tp;

	if ( msg.msg_queue_id >= MAX_QUEUES )
		return ( NULL );

	for ( op1 = task_open_table[ pid & 0xff ]; op1 != NULL;
													op1 = op1->next_open ) {
		if ( op1->open_id == msg.msg_queue_id ) {
			tp = &queue_table[ op1->open_id ];
			this_op = op1;
			if ( tp->queue_name[0] )
				return ( tp );
			}
		}

	return ( NULL );
}



int add_wait( struct queue_table_entry *tp, unsigned wait_type, int wait_proxy )
{
	struct wait_entry *wp, *wp2, *new_wp;
	unsigned seconds;

	/*
	 * Add to end of link list of wait requests.
	 */
	wp2 = NULL;
	for ( wp = tp->wait_list; wp != NULL; wp2 = wp, wp = wp->next_wait )
		/* don't allow more than one per task */
		if ( wp->wait_pid == pid )
			return ( OK );
	wp = wp2;

	/*
	 * Only do malloc when we're sure we need to.
	 */
	new_wp = malloc( sizeof( struct wait_entry ) );
	if ( new_wp == NULL )
		return ( QSTAT_MEM );

	new_wp->next_wait = NULL;
	new_wp->next_twait = NULL;
	new_wp->wait_queue_ptr = tp;
	if ( wp != NULL )
		wp->next_wait = new_wp;
	else
		tp->wait_list = new_wp;
	new_wp->wait_pid = pid;
	new_wp->wait_type = wait_type;
	new_wp->wait_size = msg.msg_queue_size;
	new_wp->wait_proxy = wait_proxy;

	/*
	 * If a timeout (in seconds) was requested we add to timer list.
	 */
	if(seconds = (wait_type >> 8) & 0xff) {
#ifdef DEBUG
    printf("seconds = %d\n",seconds);
#endif
		if((new_wp->next_twait = timer_queue) == NULL)
			settimer(1);
		else
			settimer(0);
		new_wp->wait_time = seconds;
		new_wp->wait_type |= WAIT_TIMER;
		timer_queue = new_wp;
		}
	return ( OK );
}

void close_all( void )
{
	struct open_entry *op;

	while ( (op = task_open_table[pid & 0xff]) != NULL ) {
#ifdef DEBUG
        printf("closing queue for pid %d\n",pid);
#endif
		msg.msg_queue_id = op->open_id;
		queue_close();
		}	/* while ( op != NULL ) */
}



void unlink_open( struct open_entry *entry, int id )
{
	struct open_entry **op1, **op2;

	op1 = &task_open_table[ id ];
	op2 = NULL;
	while ( *op1 != entry && *op1 != NULL ) {
		op2 = op1;
		op1 = &((*op1)->next_open);
		}
	if ( *op1 == NULL )
		/* should never happen... */
		return;

	if ( op2 != NULL )
		(*op2)->next_open = (*op1)->next_open;
	else
		/* first open entry in the list */
		task_open_table[ id ] = (*op1)->next_open;
}



void unlink_wait( struct wait_entry **list, struct wait_entry *member )
{
	struct wait_entry *wp1, *wp2;

	wp2 = NULL;
	for ( wp1 = *list; wp1 != NULL; wp2 = wp1, wp1 = wp1->next_wait ) {
		if ( wp1 == member ) {
			if ( wp2 != NULL )
				wp2->next_wait = wp1->next_wait;
			else
				/* first wait entry in the list */
				*list = wp1->next_wait;
			return;
			}
		}
}



void unlink_twait( struct wait_entry **list, struct wait_entry *member )
{
	struct wait_entry *wp1, *wp2;

	wp2 = NULL;
	for ( wp1 = *list; wp1 != NULL; wp2 = wp1, wp1 = wp1->next_twait ) {
		if ( wp1 == member ) {
			if ( wp2 != NULL )
				wp2->next_twait = wp1->next_twait;
			else
				/* first wait entry in the list */
				*list = wp1->next_twait;
			return;
			}
		}
}



void check_timer( void )
{
	struct wait_entry *wp1, *wp2;

	wp2 = NULL;
	wp1 = timer_queue;
	while ( wp1 != NULL ) {
		if ( --wp1->wait_time == 0 ) {

			msg.msg_type = QSTAT_WAITT;
			if ( wp1->wait_proxy )  {
				status_save[ wp1->wait_pid & 0xff ] = msg.msg_type;
				Trigger( wp1->wait_proxy );
				}
			else
				Reply( wp1->wait_pid, &msg, 1 );

			unlink_wait( &wp1->wait_queue_ptr->wait_list, wp1 );

			if ( wp2 != NULL )
				wp2->next_twait = wp1->next_twait;
			else
				/* first wait entry in the list */
				timer_queue = wp1->next_twait;
				
			free( wp1 );
			wp1 = (NULL == wp2) ? timer_queue : wp2->next_twait;
			}
		else {
			wp2 = wp1;
			wp1 = wp1->next_twait;
			}
		}

	if(timer_queue)
		settimer(1);
	else
		settimer(0);
}

settimer(start)
int start;
	{
	static timer_active = 0;
    struct itimerspec timer;

#ifdef DEBUG
printf("setting timer to %d\n",start);
#endif
	if(timer_active == start)
		return;
	timer_active = start;

    timer.it_value.tv_sec     = start ? 1L : 0L;
    timer.it_value.tv_nsec    = 0L;
    timer.it_interval.tv_sec  = start ? 1L : 0L;
    timer.it_interval.tv_nsec = 0L;

#ifdef DEBUG
printf("starting reltimer\n",start);
#endif
    reltimer( timerid, &timer, NULL );
	}

void queue_exit() {
	pid_t pid;
	unsigned short msg = EOK;

	qnx_pflags(0, _PPF_INFORM, 0, 0);
	while((pid = Creceive(0, 0, 0)) != -1  &&  pid != 0)
		Reply(pid, &msg, sizeof(msg));
	exit(0);
	}

