
/* trigq :  ask to be notified when an event occurs at a queue */

#ifdef __USAGE
%C	[options]

Where [options] can be:
   -q  queuename.     Open this queue name for triggering.
                      e.g. -q //0/test    Opens queue 'test' on current node.
                           -q test        Opens queue 'test' on global queue
                                          administrator (if it exists).
                           -q //23/data1  Opens queue 'data1' on node 23.
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/proxy.h>
#include <sys/kernel.h>
#include "queue.h"

struct template {
    char qheader[QHDR_SIZE];
    char data[100];
    } msg;

main(argc, argv)
int argc;
char *argv[];
{
int qid, ret;
int opt;
char queuename[80];
char retstring[40];
unsigned buf;
int status;
pid_t proxy, pid;
int wait_bits = 0x0001;   /* trigger me when the queue has a message */

    queuename[0] = NULL;

    while ((opt = getopt(argc, argv, "q:")) != -1) {
        switch(opt) {
        case 'q':
            strcpy(queuename,optarg);
            break;
        }
    }

proxy = qnx_proxy_attach( 0,0,0,-1 );
if ( proxy == -1 ) {
   printf("Can't get a proxy. Aborting.\n");
   exit(-1);
   }

/* open a queue for reading */

qid = queue_open(queuename, "r", 1000);
printf("Opening queue: %s for analysis.\n", queuename);
queue_return_code_string(qid, &retstring);
printf("Return on queue open is %d: %s\n",qid, retstring);

if ( qid < 0 )    /* can't open queue for reading so abort */
   exit(-1);

queue_wait_sig(qid, wait_bits, proxy);

while (1) {
   printf("Waiting for message or proxy trigger\n");
   pid = Receive(0, &buf, 1);
   if(pid == proxy) {  /* something happened on the queue */
      status = queue_last_status(qid); /* check to see what it was */
      queue_return_code_string(status,&retstring);
      printf("status on queue=%d:%s\n", status, retstring);
      if ( status == QSTAT_WAITM ) {
         ret = queue_read(qid, 0x0501, &msg, 100);
         if ( ret < 0 ) {
            queue_return_code_string(ret, &retstring);
            printf("Read failed. Return code=%d: %s\n",ret, retstring);
            }
         else
            printf("Got a message!: MSG=%s\n", msg.data);
         }
      queue_wait_sig(qid, wait_bits, proxy);
      }
   else {
      printf("Got message from pid %d\n",pid);
      Reply(pid, &buf, 1);
      }
   }
}

